/*
 access_log squid.
  ,       
""   ,     
  .    365/366  stat.MM.DD  ,
    dir_for_save_stat  :

IP1      
IP2      
...

:
stat_ac -l access_log_full_patch -s dir_for_save_stat
	 [-n full_patch_file_with_urls]

  -n full_patch_file_with_urls ,  
    ,     CODE/HOST,
     .    
 squid: NONE/.     bash/ksh
( [!?@*+](patlist|pat) ).
   full_patch_file_with_urls:

# ,   

DIRECT/127.0.0.+([0-9])
DIRECT/?(+([!.]).)mydomain.ru

    ,  -    
 : mydomain.ru www.mydomain.ru,   c1.clients.mydomain.ru ...
     ,   '?'  '*'.
        ,
     : DIRECT/*

: Vladimit Oleynik (C) 1999-2003 <dzo@simtreas.ru>

 *match()
    -- ksh/bash-like extended pattern matching for the shell and filename
		globbing.
   Copyright (C) 1991, 1997 Free Software Foundation, Inc.
   This function is part of GNU Bash, the Bourne Again SHell.
*/


/*     */
static char *skip_lines_in_log[] = {
  "NONE/-", /* "DIRECT/*", */
  0
};

/*        
          */
#define MAX_HOST_SAVE_IN_MEMORY 512

/* **************************************************************** */

#include <string.h>
#include <unistd.h>
#include <time.h>
#include <netdb.h>
#include <sys/socket.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>

#include "../km.h"

static int
scan_ip (char **ep, unsigned int *ip)
{
  int i, j;
  char *sp;
  unsigned char *p;
  char ip1[4];

  p = (unsigned char *) *ep;
  *ip = 0;
  for (j = 0; j < 4; j++) {
	if (*p < '0' || *p > '9')
	  return -1;
	i = 0;
	sp = ip1;
	while (*p >= '0' && *p <= '9' && i < 3) {
	  *sp = *p;
	  sp++;
	  p++;
	  i++;
	}
	if (j != 3 && *p != '.')
	  return -1;
	if (j != 3)
	  p++;
	*sp = 0;
	i = atoi (ip1);
	if (i < 0 || i > 255)
	  return -1;
	*ip = ((*ip) << 8) | i;
  }
  if (*p == 0)
	return 0;
  if (*p != '/' && *p != ' ')
	return -1;
  *ep = p;
  return 1;
}

static void *
x_malloc (unsigned int u)
{
  void *pa;

  pa = malloc (u);
  if (pa == 0) {
	fprintf (stderr, "Not enough memory\n");
	exit(2);
  }
  return pa;
}

static char *
x_strdup (const char *s)
{
  char *d;

  d = strdup (s);
  if (d == 0) {
	fprintf (stderr, "Not enough memory\n");
	exit(4);
  }
  return d;
}

static char *concat_path_file(const char *path, const char *filename)
{
	char *outbuf;
	const char *lc;

	if (!path)
	    path="";
	lc = strrchr(path, '/');
	if(lc && lc[1] != 0)
	    lc = NULL;
	while (*filename == '/')
		filename++;
	outbuf = x_malloc(strlen(path) + strlen(filename) + 2);
	sprintf(outbuf, "%s%s%s", path, (lc == NULL ? "/" : ""), filename);
	return outbuf;
}

static FILE * x_fopen(const char *name, const char *mode, const char *add_msg)
{
    FILE *f;

    f = fopen(name, mode);
    if(f == NULL) {
	fprintf(stderr, "Can`t %s %s %s: %m\n",
		(*mode == 'r' ? "open" : "create"), name, add_msg);
	exit(1);
    }
    return f;
}


char *
put_str_ip (unsigned int ip)
{
  int ip1, ip2, ip3, ip4;
  static char str[16];

  ip1 = ip & 0xff;
  ip2 = (ip & 0xff00) >> 8;
  ip3 = (ip & 0xff0000) >> 16;
  ip4 = (ip & 0xff000000) >> 24;
  sprintf (str, "%d.%d.%d.%d", ip4, ip3, ip2, ip1);
  return str;
}

struct IP_SYM_CLIENTS
{
  unsigned ip;
  double traf;
  struct IP_SYM_CLIENTS *next;
};

static struct IP_SYM_CLIENTS *top;


static void
add_traf (double traf, unsigned ip)
{
  struct IP_SYM_CLIENTS *cur;

  for (cur = top; cur != 0; cur = cur->next) {
	if (cur->ip == ip) {
	  cur->traf += traf;
	  return;
	}
  }
  cur = x_malloc (sizeof (struct IP_SYM_CLIENTS));
  cur->traf = traf;
  cur->ip = ip;
  cur->next = top;
  top = cur;
}


struct MASKS
{
  const char *m;
  struct MASKS *next;
};

static struct MASKS *mtop;

static void
load_masks (const char *path_msks)
{
    FILE *in;
    char buf[128];
    char *s, *e;
    int line = 0;
    struct MASKS *cur;

    if(!path_msks)
	return;
    in = x_fopen(path_msks, "r", "masks file");
    while( (s = fgets(buf, sizeof(buf), in)) != NULL) {
	line++;
	while(*s == ' ' || *s == '\t')
	    s++;
	if(*s == '#' || *s == '\n' || *s == 0)
	    continue;
	e = s;
	while(*++e != 0 && *e != '\n' && *e != ' ' && *e != '\t' && *e != '#')
	    ;
	*e = 0;
	e = s;
	while(*e != 0 && *e != '/')
	    e++;
	if(*e != '/' || e == s) {
	    fprintf(stderr, "Strange format masks file: line %d\n", line);
	    continue;
	}
	while(*e) {
	    if(*e >= 'a' && *e <= 'z')
		*e = *e - 'a' + 'A';
	    e++;
	}
	cur = x_malloc(sizeof(struct MASKS));
	cur->m = x_strdup(s);
	cur->next = mtop;
	mtop = cur;
    }
}

static int
load_access_log (const char *path_access_log)
{
  FILE *in;
  char *ps;
  char buf[4096];
  char *hosts[MAX_HOST_SAVE_IN_MEMORY];
  int nhosts = 0;
  unsigned ip;
  double traf;
  char *s;
  int i;
  struct MASKS *cur;

  in = x_fopen (path_access_log, "r", "log file");
  while (fgets (buf, 4096, in) != NULL) {
	ps = buf;
	s = ps;
	strtoul (s, &ps, 10);        /* time */
	if (s == ps) {
	lbf:
	  fprintf (stderr, "Bad format log file %s\n", path_access_log);
	  return 2;
	}
	if (*ps++ != '.')
	  goto lbf;
	s = ps;
	strtoul (s, &ps, 10);        /* .ms */
	if (s == ps)
	  goto lbf;
	s = ps;
	strtoul (s, &ps, 10);        /* ms */
	if (s == ps)
	  goto lbf;
	if (*ps++ != ' ')
	  goto lbf;
	if (*ps < '1' || *ps > '9')
	  goto lbf;
	s = ps;
	if (scan_ip (&ps, &ip) <= 0)
	  goto lbf;                                     /* ip */
	while (*ps == ' ')
	  ps++;
	while (*ps != ' ' && *ps != 0)
	  ps++;                         /* SQUID_CODE/HTTP_CODE */
	if (*ps++ != ' ')
	  goto lbf;
	while (*ps == ' ')
	  ps++;
	if (*ps < '0' || *ps > '9')
	  goto lbf;
	traf = strtod (ps, &ps);        /* bytes */
	if (*ps != ' ')
	  goto lbf;
	while (*ps == ' ')
	  ps++;
	while (*ps != ' ' && *ps != 0)
	  ps++;                         /* GET */
	if (*ps != ' ')
	  goto lbf;
	while (*ps == ' ')
	  ps++;
	while (*ps != ' ' && *ps != 0)
	  ps++;                         /* URL */
	if(*ps == 0) {
		/* bad url, squid error */
		if(fgets (buf, 4096, in) == NULL)
			break;
		continue;
	}
	if (*ps != ' ')
	  goto lbf;
	while (*ps == ' ')
	  ps++;
	while (*ps != ' ' && *ps != 0)
	  ps++;                         /* IDENT */
	if (*ps != ' ')
	  goto lbf;
	while (*ps == ' ')
	  ps++;
	/* toupper */
	for(s = ps; *s != 0 && *s != ' '; s++)
	    if(*s >= 'a' && *s <= 'z')
		*s = *s - 'a' + 'A';
	if (*s != ' ')
	  goto lbf;
	*s = 0;
	for (i = 0; skip_lines_in_log[i]; i++)
	  if (strcmp (ps, skip_lines_in_log[i]) == 0)
		break;
	if (skip_lines_in_log[i])
	  continue;
	for(i = 0; i < nhosts; i++) {
	    if(strcmp(ps, hosts[i]) == 0)
		break;
	}
	if(i < nhosts)
	    continue;
	for(cur = mtop; cur; cur = cur->next) {
	    if(match (cur->m, ps) == 0)
		break;
	}
	if(cur) {
	    if(nhosts < MAX_HOST_SAVE_IN_MEMORY)
		hosts[nhosts++] = x_strdup(ps);
	    continue;
	}
	add_traf (traf, ip);
  }
  fclose (in);
  return 0;
}

static void
usage (char *argv0)
{
  char *name = argv0;

  while (*argv0 != '\0') {
	if (*argv0++ == '/')
	  name = argv0;
  }
  fprintf (stderr, "Usage %s -l access.log -s dir_for_save_stat "
			"[-n full_patch_file_with_urls]\n", name);
  exit (10);
}

int
main (int argc, char **argv)
{
  int ip;
  struct IP_SYM_CLIENTS *cur;
  time_t t;
  struct tm *ptm;
  char *mm_dd;                          /* ptr to mm.dd from prevs_out_name */
  FILE *out;
  const char *access_log = NULL;
  const char *dirout = NULL;
  const char *msks = NULL;
  char *stat_out;

  while ((ip = getopt (argc, argv, "l:s:n:")) != EOF) {
	switch (ip) {
	case 'l':
	  access_log = optarg;
	  break;
	case 's':
	  dirout = optarg;
	  break;
	case 'n':
	  msks = optarg;
	  break;
	default:
	  usage (*argv);
	}
  }
  if (argc != optind || !access_log || !dirout)
	usage (*argv);

  load_masks(msks);
  ip = load_access_log (access_log);
  if (ip)
	return ip;
  time (&t);
  ptm = localtime (&t);
  stat_out = concat_path_file(dirout, "stat.MM.DD");
  mm_dd = &stat_out[strlen (stat_out) - 5];
  sprintf (mm_dd, "%02d.%02d", ptm->tm_mon + 1, ptm->tm_mday);
  out = x_fopen (stat_out, "w", "out file");
  for (cur = top; cur != 0; cur = cur->next)
    fprintf (out, "%s %.0f\n", put_str_ip (cur->ip), cur->traf);
  return 0;
}
