/*
	-CGI  squid- access.log  
	http-   URL   
	Copyright by Vladimir Oleynik <dzo@simtreas.ru> (C) 2011
*/

#include <time.h>
#include <stdio.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#define ACCESS_LOG "/usr/local/squid/logs/access.log"
//#define ACCESS_LOG "access.log"
#define ACCESS_ADDR "89.239.133.1"

#define COLOR1 "beige"
#define COLOR2 "blanchedalmond"


static void encode(const char *s)
{
	while(*s != '\0') {
		switch(*s) {
			case '<':
				printf("&lt;");
				break;
			case '>':
				printf("&gt;");
				break;
			case '&':
				printf("&amp;");
				break;
			default:
				putchar(*s);
				break;
		}
		s++;
	}
}

static char buf[24*1024];

static void cgi_err (const char *msg, ...) __attribute__ ((noreturn));

static void cgi_err (const char *msg, ...)
{
  va_list p;
  va_start(p, msg);
  int errno_sav = errno;

  printf ("<H1></H1><p>");
  errno = errno_sav;
  vsnprintf (buf, sizeof(buf)-1, msg, p);
  va_end(p);
  encode(buf);
  puts ("</p>\n</BODY>\n</HTML>");
  exit (1);
}


/*
     IP[/MASK]    :
   10 -> 10.0.0/8, 10.1 -> 10.1.0.0/16, 1.2.3.4 -> 1.2.3.4/32,
   1.2.0.0/255.255.254.0 -> 1.2.0.0/23
     
*/
static int
scan_ip (char **ep, unsigned int *ip, unsigned char endc)
{
  char *p = *ep;
  int auto_mask = 8;
  int j;

  *ip = 0;
  for (j = 0; j < 4; j++) {
	unsigned int octet;

	if ((*p < '0' || *p > '9') && (*p != '/' || j == 0) && *p != 0)
	  return -auto_mask;
	octet = 0;
	while (*p >= '0' && *p <= '9') {
	  octet *= 10;
	  octet += *p - '0';
	  if (octet > 255)
		return -auto_mask;
	  p++;
	}
	if (*p == '.')
	  p++;
	if (*p != '/' && *p != 0)
	  auto_mask += 8;
	*ip = ((*ip) << 8) | octet;
  }
  if (*p != 0) {
	if (*p != endc)
		return -auto_mask;
	p++;
	if(*p == 0)
		return -auto_mask;
  }
  *ep = p;
  return auto_mask;
}

static int
scan_1ip (char *p, unsigned int *ip)
{
  int j;
  unsigned int octet;

  *ip = 0;
  for (j = 0 ; ; j++) {
	if (*p < '0' || *p > '9')
	  return -1;
	octet = 0;
	while (*p >= '0' && *p <= '9') {
	  octet *= 10;
	  octet += *p - '0';
	  if (octet > 255)
		return -1;
	  p++;
	}
	*ip = ((*ip) << 8) | octet;
	if(j == 3) {
		if(*p != '\0')
			return -1;
		return 0;
	}
	if(*p != '.')
		return -1;
	p++;
  }
}

static int
scan_ip_mask (char *ipm, unsigned int *ip, unsigned int *mask)
{
  int i;
  unsigned int msk;

  i = scan_ip(&ipm, ip, '/');
  if(i < 0)
	return i;
  if(*ipm) {
	const char *p = ipm;

	i = 0;
	while (*p) {
		if (*p < '0' || *p > '9') {
			if (*p == '.')
				return scan_1ip (ipm, mask);
			return -1;
		}
		i *= 10;
		i += *p - '0';
		p++;
	}
  }
  if (i > 32 || i < 0)
	  return -1;
  msk = 0x80000000;
  *mask = 0;
  while (i > 0) {
	*mask |= msk;
	msk >>= 1;
	i--;
  }
  return 0;
}

#define MAX_TOKENS 10

#define START_TIME tokens[0]
#define IP_TOKENS 2
#define IP tokens[IP_TOKENS]
#define HTTP_CODE tokens[3]  /* HTTP/CODE */
#define SZ tokens[4]
#define REQUEST tokens[5]       /* request GET/PUT/CONNECT */

static char *tokens[MAX_TOKENS];

static int parse_line(char *line, unsigned int ip, unsigned int mask)
{
	char delim;
	size_t token_size;
	int num_token = 0;

	/*     */
	while (*line) {
	  token_size = strcspn (line, " \t\n");
	  delim = line[token_size];
	  if (token_size) {
		/*     */
		if(num_token < MAX_TOKENS) {
			line[token_size] = '\0';
			tokens[num_token] = line;
			if(num_token == IP_TOKENS) {
				unsigned int t_ip;

				if(scan_1ip(line, &t_ip))
					return IP_TOKENS;
				if((t_ip & mask) != ip)
					return IP_TOKENS;
			}
		} else
			return MAX_TOKENS + 1;
		num_token++;
	  }
	  /*          */
	  line += token_size + (delim != '\0');
	}
	return num_token;
}

static char *t(const char *ts)
{
	time_t t;
	struct tm *lt;
	static char stamp[100];

	t = atol(ts);
	lt = localtime( &t );
	/* make time stamp */
	strftime( stamp, sizeof( stamp ) -1, "%d/%b/%G&nbsp;%H:%M:%S", lt);
	return stamp;
}

static const char color1[] = COLOR1;
static const char color2[] = COLOR2;
static const char *color;

#define first color == NULL

static void show_line(void)
{
	char *p, *url, *rd, *mime;

	/* 1 - time of work */

	url = tokens[6];        /* URL */

	/* 7 - ident */

	rd = tokens[8];         /* DIRECT/REDIRECT */

	mime = tokens[9];       /* MIME */

	if(first) {
		puts("<TABLE border=1>\n\
 <tr>\n  <th>Date</th>\n  <th>IP</th>\n  <th>CODE</th>\n\
  <th>SIZE</th>\n  <th>R</th>\n\
  <th>URL</th>\n  <th>RD</th>\n  <th>MIME</th>\n </tr>");
		color = color1;
	}
	printf(" <tr>\n  <td bgcolor=%s>%s</td>\n  <td bgcolor=%s>%s</td>\n", color, t(START_TIME), color, IP);

	p = strchr(HTTP_CODE, '/');
	p = p != NULL ? p+1 : HTTP_CODE;
	printf("  <td bgcolor=%s>%s</td>\n", color, p);

	printf("  <td align=right bgcolor=%s>%s</td>\n  <td bgcolor=%s>%c</td>\n  <td bgcolor=%s>", color, SZ, color, REQUEST[0], color);

	if(strlen(url)>128) {
		static char show_url[132];

		p = strncpy(show_url, url, 128);
		p[128]='.';
		p[129]='.';
		p[130]='.';
		p[131]='\0';
	} else {
		p = url;
	}
	printf("<a href=\"%s\">%s</a>", url, p);

	p = strchr(mime, '/');
	if(p != NULL)
		*p = '\0';
	if(strcmp(mime, "image") == 0) {
		printf("<br><img src=\"%s\">", url);
	} else {
		p = strchr(url, '?');
		if(p != NULL)
			*p = '\0';
		p = strrchr(url, '.');
		if(p != NULL) {
			p++;
			if(strcasecmp(p, "gif") == 0 ||
			   strcasecmp(p, "jpg") == 0 ||
			   strcasecmp(p, "jpeg") == 0 ||
			   strcasecmp(p, "png") == 0)
				printf("<br><img src=\"%s\">", url);
		}
	}
	printf("</td>\n  <td bgcolor=%s>", color);

	p = strchr(rd, '/');
	if(p != NULL) {
		printf("%c%s", *rd, p);
	} else {
		printf("%s", rd);
	}

	printf("</td>\n  <td bgcolor=%s>%s</td>\n </tr>\n", color, mime);
	color = color == color1 ? color2 : color1;
}


int main(void)
{
	FILE *in;
	char *s;
	char *q = getenv("QUERY_STRING");
	char *a = getenv("REMOTE_ADDR");
	unsigned int ip, mask;
	time_t now;

	/*    */
	time(&now);
	printf ("Content-type: text/html\n\n<HTML>\n<head>\n\
<meta http-equiv=\"content-type\" content=\"text/html; charset=koi8-r\">\n\
<style>\ntable td,th {font-size: 9px; padding:0.3em;vertical-align:text-top;}\n</style>\n\
<title>Show access.log %s</title>\n</head>\n\n<BODY>\n", ctime(&now));

	if(q == NULL || *q == '\0' || scan_ip_mask(q, &ip, &mask))
		cgi_err("  URL,   ?IP  ?IP/MASK");

	if(a == NULL)
		cgi_err("  REMOTE_ADDR");
	if(strcmp(a, ACCESS_ADDR))
		cgi_err(" ");

	in = fopen(ACCESS_LOG, "r");
	if(in == NULL)
		cgi_err("     %s: %m", ACCESS_LOG);

	while ( (s = fgets(buf, sizeof(buf), in)) != NULL) {
		if(parse_line(s, ip, mask) == MAX_TOKENS)
			show_line();
	}

	puts(first ? "<h2>   IP  </h2>" : "</TABLE>");
	puts ("\n</BODY>\n</HTML>");
	return 0;
}
