/*
     squid.
  . (C) 2001-2009 <dzo@simtreas.ru>
 2.4.3


 mk_nets (c) Anatoly A. Orehovsky. AO9-RIPE. AAO1-RIPN

 *match()
    -- ksh/bash-like extended pattern matching for the shell and filename
		globbing.
   Copyright (C) 1991, 1997 Free Software Foundation, Inc.
   This function is part of GNU Bash, the Bourne Again SHell.

*/

/*   squid     -   
          */

#define CHDIR_NAME       "/usr/local/squid/etc/"
#define REDIR_CFG_NAME   "redir.cfg"
#define REDIR_ERROR_NAME "redir_error.log"
#define REDIR_LOG_NAME   "redir.log"

/*  ,     arp (Linux only?) */
#define ARP_DAT  1

#ifdef ARP_DAT
static const char arp_dat[]="/proc/net/arp";
#endif

/*  ()        acl   */
#define ALARM_TIME       20
/*        
     %hex_code */
#define PERCENT_HEX_HIDE 1

#define PREFIX_HTTP      "HTTP://"
#define PREFIX_HTTP_LEN  7
#define PREFIX_CONNECT   "CONNECT:"
#define PREFIX_CONNECT_LEN  8
#define PREFIX_FTP       "FTP://"
#define PREFIX_FTP_LEN   6
#define H_F_TP_MASK      "@(HT|F)TP://"
#define H_F_TP_MASK_LEN   12
#define DEFAULT_PREFIX    PREFIX_HTTP
#define DEFAULT_PREFIX_LEN      PREFIX_HTTP_LEN


#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/file.h>

#include <stdarg.h>

#ifdef ARP_DAT
#include <stdio.h>
#endif

#include "km.h"


/* #define DEBUG 1 */

static void
print (const char *fmt, ...)
{
  int fd, l;
  va_list args;
  const char *f;
  char *str;

  va_start (args, fmt);
  fd = open (REDIR_ERROR_NAME, O_WRONLY | O_CREAT | O_APPEND, 0660);
  if (fd > 0) {
	flock (fd, LOCK_EX);    /* squid started not 1 redirector process */
	for (;;) {
	  for (f = fmt; *f != '%' && *f != 0; ++f);
	  if (f > fmt) {
		l = f - fmt;
		write (fd, fmt, l);
		fmt = f;
	  }
	  if (*fmt == 0)
		break;
	  fmt++;
	  str = va_arg (args, char *);
	  l = strlen (str);
	  write (fd, str, l);
	}
	write (fd, "\n", 1);
	close (fd);
  }
}

static void *
my_alloc (unsigned int u)
{
  void *pa;

  pa = malloc (u);
  if (pa == 0) {
	print ("Not enough memory");
	exit (1);
  }
  return pa;
}

static
unsigned char code[256] = {     /*    glob_mask */
  0, 1, 2, 3, 4, 5, 6, 7,
  8, 9, 10, 11, 12, 13, 14, 15,
  16, 17, 18, 19, 20, 21, 22, 23,
  24, 25, 26, 27, 28, 29, 30, 31,
  ' ', '!' | 0200, '"', '#', '$', '%', '&', '\'',
  '(' | 0200, ')' | 0200, '*' | 0200, '+' | 0200, ',', '-', '.', '/',
  '0', '1', '2', '3', '4', '5', '6', '7',
  '8', '9', ':', ';', '<', '=', '>', '?' | 0200,
  ' ' | 0200, 'A', 'B', 'C', 'D', 'E', 'F', 'G',
  'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
  'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
  'X', 'Y', 'Z', '.' | 0200, '\\', ',' | 0200, '-' | 0200, '_',
  '`', 'A', 'B', 'C', 'D', 'E', 'F', 'G',
  'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
  'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
  'X', 'Y', 'Z', '{', '|', '}', '~', 127,
  128, 129, 130, 131, 132, 133, 134, 135,
  136, 137, 138, 139, 140, 141, 142, 143,
  144, 145, 146, 147, 148, 149, 150, 151,
  152, 153, 154, 155, 156, 157, 158, 159,
  160, 161, 162, 163, 164, 165, 166, 167,
  168, 169, 170, 171, 172, 173, 174, 175,
  176, 177, 178, 179, 180, 181, 182, 183,
  184, 185, 186, 187, 188, 189, 190, 191,
  192, 193, 194, 195, 196, 197, 198, 199,
  200, 201, 202, 203, 204, 205, 206, 207,
  208, 209, 210, 211, 212, 213, 214, 215,
  216, 217, 218, 219, 220, 221, 222, 223,
  224, 225, 226, 227, 228, 229, 230, 231,
  232, 233, 234, 235, 236, 237, 238, 239,
  240, 241, 242, 243, 244, 245, 246, 247,
  248, 249, 250, 251, 252, 253, 254, 255,
};




static int
scan_ip (char **ep, unsigned int *ip)
{
  int i, j;
  unsigned char *p;
  int ip1;

  p = (unsigned char *) *ep;
  *ip = 0;
  for (j = 0; j < 4; j++) {
	if (*p < '0' || *p > '9')
	  return -1;
	i = 0;
	ip1 = 0;
	while (*p >= '0' && *p <= '9' && i < 3) {
	  ip1 *= 10;
	  ip1 += *p - '0';
	  p++;
	  i++;
	}
	if (j != 3 && *p != '.')
	  return -1;
	if (j != 3)
	  p++;
	if (ip1 > 255)
	  return -1;
	*ip = ((*ip) << 8) | ip1;
  }
  if (*p == 0)
	return 0;
  if (*p != '/')
	return -1;
  p++;
  *ep = (char *) p;
  return 1;
}

static int
scan_mask (char **ep, unsigned int *ip)
{
  int i;
  unsigned char *p;

  p = (unsigned char *) *ep;
  i = 0;
  while (*p) {
	if (*p < '0' || *p > '9') {
	  if (*p++ == '.')
		return scan_ip (ep, ip);
	  return -1;
	}
	i *= 10;
	i += *p - '0';
	p++;
  }
  *ep = (char *) p;
  if (i > 32 || i < 0)
	return -1;
  if(i == 32)
	*ip = 0xffffffff;
  else
	*ip = ~ (0xffffffffUL >> i);
  return 0;
}

struct TIME_DECODE
{
  int d1;
  int d2;
  long min1;
  long min2;
};
static struct TIME_DECODE T_DECODE_all[1] = { {0, 6, 0, 24l * 60 * 60} };

struct ALC_STRUCT
{
  int type;
  unsigned ip;
  unsigned mask;
  int ntime;
  struct TIME_DECODE *time;
  int log_type;
};
static struct ALC_STRUCT ALC_STRUCT_all = { 'R', 0, 0, 1, T_DECODE_all, 0 };

struct ALC_COMPLEX
{
  int n;
  struct ALC_STRUCT *dim_acl;
  struct ALC_COMPLEX *next;
  char name[1];                 /*   */
};
static struct ALC_COMPLEX ALC_COMPLEX_all = { 1, &ALC_STRUCT_all, 0, "*" };

static struct ALC_COMPLEX *top_acl_com = &ALC_COMPLEX_all;


/* format: * or 0-6:1-12.59,1:* */

static long
decode_hhmm (const char **pst)
{
  long m;
  const char *s;

  s = *pst;
  m = strtol (s, (char **)pst, 10);
  if (*pst == s || m >= 24 || m < 0)
	return -1;
  m *= 60;
  if (**pst == '.') {
	(*pst)++;
	s = *pst;
	m += strtol (s, (char **)pst, 10);
	if (*pst == s || m > (24 * 60) || m < 0)
	  return -1;
	if (m == (24 * 60))
	  m--;
  }
  return m * 60;
}

static int
decode_time (const char *time, int *ntime, struct TIME_DECODE **ptd)
{
  const char *s;
  int n;
  struct TIME_DECODE *p;

  n = 1;
  s = time;
  while (*s) {
	if (*s == ',')
	  n++;
	s++;
  }
  *ntime = n;
  p = my_alloc (n * sizeof (struct TIME_DECODE));
  *ptd = p;
  s = time;
  for (n = 0;; n++) {
	if (*s == '*' && (*(s + 1) == 0 || *(s + 1) == ':' || *(s + 1) == '-')) {
	  p[n].d1 = 0;
	  p[n].d2 = 6;
	  if (*(s + 1) == 0) {
		p[n].min1 = 0;
		p[n].min2 = 24l * 60 * 60;
		break;
	  } else
		s++;
	} else {
	  if (*s < '0' || *s > '6')
		return 1;
	  p[n].d1 = *s++ - '0';
	}
	if (*s == '-') {
	  s++;
	  if (*s < '0' || *s > '6')
		return 1;
	  p[n].d2 = *s++ - '0';
	} else if (*(s - 1) != '*')
	  p[n].d2 = p[n].d1;
	if (p[n].d2 < p[n].d1) {
	  int swd;
	  swd = p[n].d1;
	  p[n].d1 = p[n].d2;
	  p[n].d2 = swd;
	}
	if (*s != ':')
	  return 1;
	s++;
	if (*s == '*') {
	  s++;
	  p[n].min1 = 0;
	  p[n].min2 = 24l * 60 * 60;
	} else {
	  p[n].min1 = decode_hhmm (&s);
	  if (p[n].min1 < 0)
		return 1;
	  if (*s == 0) {
		p[n].min2 = p[n].min1;
		break;
	  }
	  if (*s == ',') {
		p[n].min2 = p[n].min1;
		s++;
		continue;
	  }
	  if (*s++ != '-')
		return 1;
	  if (*s == '*') {
		s++;
		p[n].min2 = (23l * 60 + 59) * 60;
	  } else {
		p[n].min2 = decode_hhmm (&s);
		if (p[n].min2 < 0)
		  return 1;
	  }
	}
	if (*s == 0)
	  break;
	if (*s++ != ',')
	  return 1;
  }
  return 0;
}

static int mk_nets_count;
static void
mk_nets (unsigned long f, unsigned long l, struct ALC_STRUCT *dim_acl)
{
  unsigned long mask;

  mask = 0xffffffffL;

  while (((f & mask) != (l & mask)) && ((f & (mask << 1)) == f) &&
		 ((f | ~(mask << 1)) <= l))
	mask <<= 1;
  if (dim_acl) {
	dim_acl[mk_nets_count].ip = f & mask;
	dim_acl[mk_nets_count].mask = mask;
  }
  mk_nets_count++;
  if ((f & mask) != (l & mask))
	mk_nets ((f | ~mask) + 1, l, dim_acl);

}


#ifdef ARP_DAT
static int
decode_arp (const char *sarp, char arpd[6])
{

  int i, c, c2;

  for (i = 0; i < 6; i++) {
	c = *sarp++;
	if (c >= 'A' && c <= 'F')
	  c -= 'A' - 'a';
	if (c >= '0' && c <= '9')
	  c = c - '0';
	else if (c >= 'a' && c <= 'f')
	  c = c - 'A' + 10;
	else
	  return -1;
	c2 = *sarp++;
	if (i == 5) {
	  if (c2 == ':')
		return -1;
	  if (c2 == 0 || c2 == ' ' || c2 == '\t' || c2 == '\n')
		c2 = ':';
	}
	arpd[i] = c;
	if (c2 != ':') {
	  arpd[i] *= 16;
	  c = c2;
	  if (c >= 'A' && c <= 'F')
		c -= 'A' - 'a';
	  if (c >= '0' && c <= '9')
		c = c - '0';
	  else if (c >= 'a' && c <= 'f')
		c = c - 'A' + 10;
	  else
		return -1;
	  arpd[i] += c;
	  c2 = *sarp++;
	}
	if (i == 5) {
	  if (c2 == ':')
		return -1;
	  if (c2 == 0 || c2 == ' ' || c2 == '\t' || c2 == '\n')
		c2 = ':';
	}
	if (c2 != ':')
	  return -1;
  }
  return 0;
}
#endif


#define LOG_OPT_L 1                     /* acl type L */

#define LOG_OPT_D 2
#define LOG_OPT_U 4
#define LOG_OPT_C 8
#define LOG_OPT_G 16
#define LOG_OPT_N 32
#define LOG_OPT_O 64
#define LOG_OPT_R 128
#define LOG_OPT_M 256

#define LOG_OPT_TD 512                  /* acl action D */
#define LOG_OPT_TR 1024                 /* acl action R */
#define LOG_OPT_TA 2048                 /* acl action * */

#define LOG_OPT_P  4096                 /* acl type P */

static int
acl_add (int type, const char *name, char *ip_mask, const char *time)
{
  struct ALC_COMPLEX *cur;
  struct ALC_COMPLEX *cur2;
  char *ps;
  char **names;
  int n, i;

  for (cur2 = top_acl_com; cur2 != 0; cur2 = cur2->next)
	if (!strcmp (name, cur2->name)) {
	  print ("Acl % allready", name);
	  return -1;
	}
  cur = my_alloc (sizeof (struct ALC_COMPLEX) + strlen(name));
  cur->next = 0;
  n = 1;
  ps = ip_mask;
  while (*ps) {
	if (*ps == '+')
	  n++;
	ps++;
  }
  names = (char **) alloca (n * sizeof (char *));
  ps = ip_mask;
  for (i = 0; *ps; i++) {
	names[i] = ps;
	while (*ps) {
	  if (*ps == '+') {
		*ps = 0;
		ps++;
		break;
	  }
	  ps++;
	}
  }
  cur->n = 0;
  for (i = 0; i < n; i++) {
	ps = names[i];
	while (*ps) {
	  if (*ps == '/') {
		cur->n++;
		break;
	  }
	  if (*ps == '-') {
		char *s2;
		unsigned int fi, li;

		*ps++ = 0;
		s2 = names[i];
		if (scan_ip (&s2, &fi) != 0)
		  goto err;
		if (scan_ip (&ps, &li) != 0)
		  goto err;
		if (fi > li) {
		  unsigned t;
		  t = fi;
		  fi = li;
		  li = t;
		}
		mk_nets_count = 0;
		mk_nets (fi, li, 0);
		if (mk_nets_count <= 0)
		  goto err;
		cur->n += mk_nets_count;
		*--ps = '-';
		break;
	  }
	  ps++;
	}
	if (*ps == 0) {   /* IP without "/mask" */
	  char *s2 = names[i];
	  unsigned int fi;

	  if(scan_ip (&s2, &fi) != 0) {
err:
	    print ("Bad format IP % while setup acl %", names[i], name);
	    return 1;
	  }
	  cur->n++;
	}
  }
  cur->dim_acl = my_alloc (cur->n * sizeof (struct ALC_STRUCT));
  cur->n = 0;
  for (i = 0; i < n; i++) {
	ps = names[i];
	while (*ps) {
	  if (*ps == '/') {
		ps = names[i];
		if (scan_ip (&ps, &(cur->dim_acl[cur->n].ip)) <= 0)
		  goto err;
		if (scan_mask (&ps, &(cur->dim_acl[cur->n].mask)) != 0) {
		  print ("Bad MASK (%) while setup acl %", ps, name);
		  return 1;
		}
		cur->n++;
		ps = 0;
		break;
	  }
	  if (*ps == '-') {
		char *s2;
		unsigned int fi, li;
		*ps++ = 0;
		s2 = names[i];
		scan_ip (&s2, &fi);
		scan_ip (&ps, &li);
		if (fi > li) {
		  unsigned t;
		  t = fi;
		  fi = li;
		  li = t;
		}
		mk_nets_count = cur->n;
		mk_nets (fi, li, cur->dim_acl);
		cur->n = mk_nets_count;
		ps = 0;
		break;
	  }
	  ps++;
	}
	if(ps!=0) {   /* IP without "/mask" */
	  ps = names[i];
	  scan_ip (&ps, &(cur->dim_acl[cur->n].ip));
	  cur->dim_acl[cur->n].mask = 0xffffffffL;  /* mask = "/32" */
	  cur->n++;
	}
  }
  cur2 = top_acl_com;
  while (cur2->next != 0)
	cur2 = cur2->next;
  cur2->next = cur;
  strcpy (cur->name, name);
  if (decode_time (time, &cur->dim_acl[0].ntime, &cur->dim_acl[0].time)) {
	print ("Bad time format while setup acl %", name);
	return -1;
  }
  for (i = 0; i < cur->n; i++) {
	if (i) {
	  cur->dim_acl[i].time = cur->dim_acl[0].time;
	  cur->dim_acl[i].ntime = cur->dim_acl[0].ntime;
	  cur->dim_acl[i].log_type = LOG_OPT_P;
	} else
	  cur->dim_acl[i].log_type = 0;
	cur->dim_acl[i].type = type;
  }
  return 0;
}



static int
set_log_acl (const char *name, char *type, char *opt)
{
  struct ALC_COMPLEX *cur;
  struct ALC_COMPLEX *cur2;
  int c;

  for (cur2 = top_acl_com; cur2 != 0; cur2 = cur2->next)
	if (!strcmp (name, cur2->name)) {
	  print ("Acl % allready", name);
	  return -1;
	}
  cur = my_alloc (sizeof (struct ALC_COMPLEX) + strlen(name));
  cur->dim_acl = my_alloc (sizeof (struct ALC_STRUCT));
  cur->next = 0;
  cur->n = 1;
  cur->dim_acl[0].log_type = LOG_OPT_L;
  while (*type) {
	c = *type;
	c &= 0337;
	switch (c) {
	case 'D':
	  cur->dim_acl[0].log_type |= LOG_OPT_TD;
	  break;
	case 'R':
	  cur->dim_acl[0].log_type |= LOG_OPT_TR;
	  break;
	case ('*' & 0337):
	  cur->dim_acl[0].log_type |= LOG_OPT_TA;
	  break;
	default:
	  *(type + 1) = 0;
	  print ("L %: bad action %", name, type);
	  return -1;
	}
	type++;
  }
  while (*opt) {
	c = *opt;
	c &= 0337;
	switch (c) {
	case 'D':
	  cur->dim_acl[0].log_type |= LOG_OPT_D;
	  break;
	case 'U':
	  cur->dim_acl[0].log_type |= LOG_OPT_U;
	  break;
	case 'C':
	  cur->dim_acl[0].log_type |= LOG_OPT_C;
	  break;
	case 'M':
	  cur->dim_acl[0].log_type |= LOG_OPT_G;
	  break;
	case 'N':
	  cur->dim_acl[0].log_type |= LOG_OPT_N;
	  break;
	case 'O':
	  if ((cur->dim_acl[0].log_type & (LOG_OPT_R | LOG_OPT_M)) != 0) {
		print ("L %: option O not compatibile witch options R or -", name);
		return -1;
	  }
	  cur->dim_acl[0].log_type |= LOG_OPT_O;
	  break;
	case 'R':
	  if ((cur->dim_acl[0].log_type & (LOG_OPT_O | LOG_OPT_M)) != 0) {
		print ("L %: option R not compatibile witch options O or -", name);
		return -1;
	  }
	  cur->dim_acl[0].log_type |= LOG_OPT_R;
	  break;
	case ('-' & 0337):
	  if ((cur->dim_acl[0].log_type & (LOG_OPT_O | LOG_OPT_R)) != 0) {
		print ("L %: option - not compatibile witch options O or R", name);
		return -1;
	  }
	  cur->dim_acl[0].log_type |= LOG_OPT_M;
	  break;
	case (',' & 0337):
	  break;
	default:
	  *(opt + 1) = 0;
	  print ("L %: bad option %", name, opt);
	  return -1;
	}
	opt++;
  }
  cur2 = top_acl_com;
  while (cur2->next != 0)
	cur2 = cur2->next;
  cur2->next = cur;
  strcpy (cur->name, name);
  return 0;
}

static int
acl_paste (const char *name, char *name_plus)
{
  struct ALC_COMPLEX *cur;
  struct ALC_COMPLEX *cur2;
  char **names;
  char *ps;
  int i, n, j;

  for (cur2 = top_acl_com; cur2 != 0; cur2 = cur2->next)
	if (!strcmp (name, cur2->name)) {
	  print ("Acl % allready", name);
	  return -1;
	}
  cur = my_alloc (sizeof (struct ALC_COMPLEX) + strlen(name));
  cur->next = 0;
  cur2 = top_acl_com;
  while (cur2->next != 0)
	cur2 = cur2->next;
  cur2->next = cur;
  strcpy (cur->name, name);
  n = 1;
  ps = name_plus;
  while (*ps) {
	if (*ps == '+')
	  n++;
	ps++;
  }
  names = (char **) alloca (n * sizeof (char *));
  ps = name_plus;
  for (i = 0; *ps; i++) {
	names[i] = ps;
	while (*ps) {
	  if (*ps == '+') {
		*ps = 0;
		ps++;
		break;
	  }
	  ps++;
	}
  }
  j = 0;
  for (i = 0; i < n; i++) {
	for (cur2 = top_acl_com; cur2 != 0; cur2 = cur2->next)
	  if (!strcmp (names[i], cur2->name))
		break;
	if (cur2 == 0) {
	  print ("Not found % acl", names[i]);
	  return -1;
	}
	if ((cur2->dim_acl[0].log_type & LOG_OPT_L) == 0)
	  j += cur2->n;
  }
  cur->n = j;
  cur->dim_acl = my_alloc (j * sizeof (struct ALC_STRUCT));
  j = 0;
  for (i = 0; i < n; i++) {
	int k;
	for (cur2 = top_acl_com; cur2 != 0; cur2 = cur2->next)
	  if (!strcmp (names[i], cur2->name))
		break;
	if ((cur2->dim_acl[0].log_type & LOG_OPT_L) == 0) {
	  for (k = 0; k < cur2->n; k++) {
		cur->dim_acl[j] = cur2->dim_acl[k];
		cur->dim_acl[j].log_type |= LOG_OPT_P;
		j++;
	  }
	} else {
	  for (k = 0; k < j; k++) {
		cur->dim_acl[k].log_type |= cur2->dim_acl[0].log_type;
		cur->dim_acl[k].log_type &= ~LOG_OPT_L;
	  }
	}
  }
  return 0;
}

struct LIST_LOCAL_URL_HOST
{
  char *value;
  int len_def;
  struct LIST_LOCAL_URL_HOST *next;
  char name_var[1];                     /*   */
};

static struct LIST_LOCAL_URL_HOST *top_lluh;

static char *
expand_var (const char *url_var, int *allocated)
{
  int i, v;
  char *var;
  char *ret_url;
  const char *sav_var;
  char null_str[2];
  char **ptr_values;
  char **vars_values;
  struct LIST_LOCAL_URL_HOST *cur;

  sav_var = url_var;
  i = 0;
  while (*sav_var)
	if (*sav_var++ == '$')
	  i++;
  if(allocated) {
	*allocated = i;
	if(i == 0)
		return (char *)url_var;
  }
  vars_values = (char **) alloca (i * sizeof (char *));
  ptr_values  = (char **) alloca (i * sizeof (char *));
  i = v = 0;
  sav_var = url_var;
  while (*sav_var) {
	if (*sav_var++ != '$')
	  i++;
	else {
	  char *tmp;

	  if (*sav_var == 0) {
		print ("Empty variable name");
		return 0;
	  }
	  var = strcpy (alloca (strlen (sav_var) + 1), sav_var);
	  tmp = var;
	  while (*++tmp)
		if (*tmp == '$' || *tmp == '/' || *tmp == '+') {
		  *tmp = 0;
		  break;
		}
	  for (cur = top_lluh; cur != 0; cur = cur->next)
		if (!strcmp (var, cur->name_var))
		  break;
	  if (cur == 0) {
		print ("Variable % not defined", var);
		return 0;
	  }
	  i += cur->len_def;
	  vars_values[v]  = cur->value;
	  ptr_values[v++] = cur->name_var;
	  sav_var += (tmp - var);
	}
  }
  null_str[0] = null_str[1] = 0;
  ret_url = strcpy (my_alloc (i + 1), null_str);
  sav_var = url_var;
  v = 0;
  while ((i = *sav_var++)) {
	if (i != '$') {
	  null_str[0] = i;
	  strcat (ret_url, null_str);
	} else {
	  strcat (ret_url, vars_values[v]);
	  sav_var += strlen (ptr_values[v++]);
	}
  }
  return ret_url;
}

static int
def_var (const char *name_var, const char *value)
{
  struct LIST_LOCAL_URL_HOST *cur;
  struct LIST_LOCAL_URL_HOST *cur2;

  if (*name_var == '$')
	name_var++;
  for (cur2 = top_lluh; cur2 != 0; cur2 = cur2->next)
	if (!strcmp (name_var, cur2->name_var)) {
	  print ("Variable % allready defined", name_var);
	  return -1;
	}
  cur = my_alloc (sizeof (struct LIST_LOCAL_URL_HOST) + strlen(name_var));
  cur->value = expand_var (value, 0);
  if (cur->value == 0)
	return -1;
  cur->len_def = strlen (cur->value);
  cur->next = top_lluh;
  strcpy (cur->name_var, name_var);
  top_lluh = cur;                       /* reverse */
  return 0;
}

struct LIST_U_RU
{
  char *url;
  char *rurl;
  int  jump_mirror_connect;
  struct ALC_COMPLEX *acl_name;
  struct LIST_U_RU *next;
};

static struct LIST_U_RU *top, *top_connect;

static char *tst_and_add_if_need_prefix(const char *u, struct LIST_U_RU *cur)
{
	char *u2;

	if(strncasecmp (u, PREFIX_CONNECT, PREFIX_CONNECT_LEN) == 0) {
		u2 = strcpy (my_alloc (strlen(u + PREFIX_CONNECT_LEN) + 1),
					   u + PREFIX_CONNECT_LEN);
		if(cur != NULL)
			cur->jump_mirror_connect = 3;
		else
			return NULL;
		return u2;
	}
	if(strncasecmp (u, PREFIX_HTTP, PREFIX_HTTP_LEN) &&
	   strncasecmp (u, PREFIX_FTP, PREFIX_FTP_LEN) &&
	   strncasecmp (u, H_F_TP_MASK, H_F_TP_MASK_LEN)) {
		u2 = strcpy (my_alloc (strlen(u) + 1 + DEFAULT_PREFIX_LEN),
					   DEFAULT_PREFIX);
		strcat (u2, u);
		return u2;
	}
	return (char *)u;
}

/*  acl_name=star,    acl_name  ru   D- */
static char star[] = "*";

static struct LIST_U_RU *
url_add (char *url, const char *ru, const char *acl_name)
{
  unsigned char *s;
  char *s0;
  struct LIST_U_RU *cur;
  struct ALC_COMPLEX *cur_a;

  if(acl_name == star && strncasecmp (ru, PREFIX_HTTP, PREFIX_HTTP_LEN) &&
		strncasecmp (ru, PREFIX_FTP, PREFIX_FTP_LEN)) {
	struct ALC_COMPLEX *rd_a;

	for (rd_a = top_acl_com; rd_a != 0; rd_a = rd_a->next)
		if (!strcmp (ru, rd_a->name))
			break;
	if(rd_a != NULL) {
		int n;

		for(n = 0; n < rd_a->n; n++)
			if(rd_a->dim_acl[n].type == 'R')
				break;
		if(n == rd_a->n) {
			/*    D,  ru -  acl */
			acl_name = ru;
		}
	}
  }
  for (cur_a = top_acl_com; cur_a != 0; cur_a = cur_a->next)
	if (!strcmp (acl_name, cur_a->name))
	  break;
  if (cur_a == 0) {
	print ("Acl % not defined", acl_name);
	return 0;
  }
  if ((cur_a->dim_acl[0].log_type & LOG_OPT_L) != 0) {
	print ("Acl % is Log-type, please paste witch R/D-type acl", acl_name);
	return 0;
  }
  cur = my_alloc (sizeof (struct LIST_U_RU));
  cur->acl_name = cur_a;
  cur->jump_mirror_connect = 0;
  cur->next = 0;
  cur->rurl = expand_var (ru, 0);
  if (cur->rurl == 0)
	return 0;
  s0 = url;
  s = (unsigned char *)url;
  while (*s) {
	if (*s == '\\') {
	  s++;
	  if(*s == '\0')
		break;
	  *s0 = code[*s];
	} else if(*s >= 'a') {
	  *s0 = code[*s];
	} else {
	  *s0 = *s;
	}
	s0++;
	s++;
  }
  *s0 = '\0';
  cur->url = tst_and_add_if_need_prefix(url, cur);
  if(cur->url == url)
	cur->url = strcpy (my_alloc ((s0 - url) + 1), url);
  if(cur->jump_mirror_connect == 3) {
	cur->jump_mirror_connect = 0;
	if (top_connect == 0)
		top_connect = cur;
	else {
		struct LIST_U_RU *cur2 = top_connect;

		while (cur2->next != 0)
			cur2 = cur2->next;
		cur2->next = cur;
	}
  } else {
	if (top == 0)
		top = cur;
	else {
		struct LIST_U_RU *cur2 = top;

		while (cur2->next != 0)
			cur2 = cur2->next;
		cur2->next = cur;
	}
  }
  return cur;
}

static int
jump_add (char *url, char *rurl, const char *acl_name)
{
  struct LIST_U_RU *cur;
  size_t l2;
  unsigned char *s;
  char *s0;

  cur = url_add(url, rurl, acl_name);
  if(cur == 0)
	return -1;
  s0 = cur->rurl;
  s = (unsigned char *)s0;
  while (*s) {
	if (*s == '\\') {
	  s++;
	  if(*s == '\0')
		break;
	  *s0 = code[*s];
	} else if(*s >= 'a') {
	  *s0 = code[*s];
	} else {
	  *s0 = *s;
	}
	s0++;
	s++;
  }
  *s0 = '\0';
  s0 = tst_and_add_if_need_prefix(cur->rurl, NULL);
  if(s0 == NULL) {
	print ("%: URL for jump must do not CONNECT", cur->rurl);
	return -1;
  }
  if(s0 != cur->rurl) {
	free(cur->rurl);
	cur->rurl = s0;
  }

  l2 = strlen(cur->rurl);
  if(strlen(cur->url) <= l2 || strncmp(cur->url, cur->rurl, l2) != 0) {
    print ("%: URL for jump mode must do partial %", cur->rurl, cur->url);
    return -1;
  }
  cur->jump_mirror_connect = 1;
  return 0;
}

static int
mirror_add (char *url, char *rurl, const char *acl_name)
{
  struct LIST_U_RU *cur;
  size_t l2;
  char *s;

  cur = url_add(url, rurl, acl_name);
  if(cur == 0)
	return -1;
  l2 = strlen(url);
  if(url[l2-1] != '*') {
	print ("%: URL for mirror mode must have tail '*'", url);
	return -1;
  }
  s = tst_and_add_if_need_prefix(cur->rurl, NULL);
  if(s == NULL) {
	print ("%: URL for mirror must do not CONNECT", url);
	return -1;
  }
  if(s != cur->rurl) {
	free(cur->rurl);
	cur->rurl = s;
  }

  cur->jump_mirror_connect = 2;
  return 0;
}


struct F_URLS_REREAD
{
  time_t st_m;
  struct F_URLS_REREAD *next;
  char name[1];                 /*   */
};

static struct F_URLS_REREAD *top_fur;

static void
furl_add (const char *name, time_t st_m)
{
  struct F_URLS_REREAD *cur;

  cur = my_alloc (sizeof (struct F_URLS_REREAD) + strlen(name));
  cur->st_m = st_m;
  cur->next = top_fur;  /* reverse */
  top_fur = cur;
  strcpy (cur->name, name);
}


#ifdef ARP_DAT
struct ARP_LIST
{
  unsigned ip;
  struct ALC_COMPLEX *log_acl;
  char arp[6];
  char *rurl;
  struct ARP_LIST *next;
};

struct ARP_LIST_CMP
{
  unsigned ip;
  char arp[6];
  struct ARP_LIST_CMP *next;
 };

static struct ARP_LIST *arp_top;
static struct ARP_LIST_CMP *arp_cmp_top;

static int
arp_add (char *ip_l, const char *arp, char *url)
{
  unsigned int fi;
  char *s;
  int i;
  struct ALC_COMPLEX *cur_a = 0;
  struct ARP_LIST *cur;
  struct ARP_LIST *b;

  s = ip_l;
  i = scan_ip (&s, &fi);
  if (i < 0) {
	print ("Bad decode ip address A %", ip_l);
	return -1;
  }
  if (i) {
	for (cur_a = top_acl_com; cur_a != 0; cur_a = cur_a->next)
	  if (!strcmp (s, cur_a->name))
		break;
	if (cur_a == 0) {
	  print ("A %: acl % not defined", ip_l, s);
	  return -1;
	}
	if ((cur_a->dim_acl[0].log_type & LOG_OPT_L) == 0) {
	  print ("A %: acl % is not Log-type", ip_l, s);
	  return -1;
	}
  }
  cur = my_alloc (sizeof (struct ARP_LIST));
  if (decode_arp (arp, cur->arp)) {
	print ("Bad decode arp A % %", ip_l, arp);
	return -1;
  }
  if(url == 0) {
	for (b = arp_top; b != 0; b = b->next)
		if (b->ip == fi)
			url = b->rurl;
	if(url) {
		cur->rurl = strcpy(my_alloc(strlen(url) + 1), url);
	} else {
		print ("Not found redirect url for A % %", ip_l, arp);
		return -1;
	}
  } else {
	cur->rurl = url;
  }
  cur->ip = fi;
  cur->log_acl = cur_a;
  cur->next = arp_top;
  arp_top = cur;        /* reverse */
  return 0;
}
#endif


static void
free_all_lists (void)
{
  void *cur;
  void *prev;

  for (cur = (void *) (top_acl_com->next); cur != 0;) {
	int i;

	i = ((struct ALC_COMPLEX *) cur)->n;
	while (--i >= 0)
	  if ((((struct ALC_COMPLEX *) cur)->dim_acl[i].log_type &
		   (LOG_OPT_P | LOG_OPT_L)) == 0)
		free (((struct ALC_COMPLEX *) cur)->dim_acl[i].time);
	free (((struct ALC_COMPLEX *) cur)->dim_acl);
	prev = cur;
	cur = (void *) (((struct ALC_COMPLEX *) cur)->next);
	free (prev);
  }
  top_acl_com->next = 0;
  for (cur = (void *) top; cur != 0;) {
	free (((struct LIST_U_RU *) cur)->url);
	free (((struct LIST_U_RU *) cur)->rurl);
	prev = cur;
	cur = (void *) (((struct LIST_U_RU *) cur)->next);
	free (prev);
  }
  top = 0;
  for (cur = (void *) top_connect; cur != 0;) {
	free (((struct LIST_U_RU *) cur)->url);
	free (((struct LIST_U_RU *) cur)->rurl);
	prev = cur;
	cur = (void *) (((struct LIST_U_RU *) cur)->next);
	free (prev);
  }
  top_connect = 0;
  for (cur = (void *) top_lluh; cur != 0;) {
	free (((struct LIST_LOCAL_URL_HOST *) cur)->value);
	prev = cur;
	cur = (void *) (((struct LIST_LOCAL_URL_HOST *) cur)->next);
	free (prev);
  }
  top_lluh = 0;
  for (cur = (void *) top_fur; cur != 0;) {
	prev = cur;
	cur = (void *) (((struct F_URLS_REREAD *) cur)->next);
	free (prev);
  }
  top_fur = 0;
#ifdef ARP_DAT
  for (cur = (void *) arp_top; cur != 0;) {
	free (((struct ARP_LIST *) cur)->rurl);
	prev = cur;
	cur = (void *) (((struct ARP_LIST *) cur)->next);
	free (prev);
  }
  arp_top = 0;
#endif
}

static void
put_timer (char *buf, time_t t)
{
  char str1[2];
  time_t r;

  if ((r = t / 10) != 0)
	put_timer (buf, r);
  str1[1] = 0;
  str1[0] = (t % 10) + '0';
  strcat (buf, str1);
}

static int
strip_space (char *s, int size)
{                               /*   -    */
  register char *s2;
  register char *s3;

  if (size <= 0)
	return 0;
  s2 = s + size;
  if (*(s2 - 1) != '\n')
	*s2++ = '\n';           /*      */
  *s2 = 0;
  s2 = s3 = s;
  while (*s2) {
	if (*s2 == '#') {       /*   */
	  while (*++s2 != '\n');
	} else if (*s2 == '\t')
	  *s3++ = ' ';            /*  \t   */
	else if (*s2 == '\\' && *(s2 + 1) == '\n') {
	  /*    : \\n */
	  s2++;
	} else
	  *s3++ = *s2;
	s2++;
  }
  *s3 = 0;
  s2 = s;                       /*  \n      */
  while (*s2 == ' ' || *s2 == '\n')
	s2++;
  s3 = s;
  while (*s2) {                 /*       */
	if (*s2 == ' ' && (*(s2 + 1) == ' ' || *(s2 + 1) == '\n')) {
	  size = *(s2 + 1) == '\n';
	  s2 += 2;
	  while (*s2 == ' ' || *s2 == '\n') {
		size |= *s2 == '\n';
		s2++;
	  }
	  *s3++ = size ? '\n' : ' ';
	} else if (*s2 == '\n' && (*(s2 + 1) == ' ' || *(s2 + 1) == '\n')) {
	  s2 += 2;
	  while (*s2 == ' ' || *s2 == '\n')
		s2++;
	  *s3++ = '\n';
	} else
	  *s3++ = *s2++;
  }
  *s3 = 0;
  return (int) ((unsigned char) *s);
}

static char *
get_parm (char *s, int delim1, int delim2, char *sdelim)
{
  while (*s != ' ' && *s != '\n' && *s != 0)
	s++;
  if (*s != delim1 && *s != delim2)
	return 0;
  if (sdelim)
	*sdelim = *s;
  *s++ = 0;
  return s;
}

static int
read_ub (const char *ncfg, int recurs)
{
  struct stat statb;
  int n, f;
  char *s, *sold, *rs, *an;
  char  delim;
  char *allocate1, *allocate2;
  int allocated;

  if (stat (ncfg, &statb) < 0) {
	print ("Can`t stat cfg_file %", ncfg);
	return 0;
  }
  s = my_alloc (statb.st_size + 2);
  allocate1 = s;
  f = open (ncfg, O_RDONLY);
  if (f < 0) {
	print ("Can`t open cfg_file %", ncfg);
	return 0;
  }
  furl_add (ncfg, statb.st_mtime);      /*     */
  n = read (f, s, statb.st_size);
  if (n != statb.st_size) {
	print ("Error reading cfg_file");
	return 0;
  }
  close (f);
  if (!strip_space (s, n)) {
  erp:
	print ("Error parsing cfg_file");
	return 0;
  }
  allocate2 = my_alloc (strlen (s) + 1);
  strcpy (allocate2, s);
  free (allocate1);
  allocate1 = s = allocate2;
  n = 0;                                                /*   */
  while (*s) {
	f = *s++;                                       /*  */
	f &= 0337;
	if (f != 'U' && f != 'F' && f != 'D' && f != 'R' && f != 'P'
		     && f != 'S' && f != 'L' && f != 'I' && f != 'A'
		     && f != 'J' && f != 'M')
	  goto erp;
	if (*s != ' ')
	  goto erp;
	sold = ++s;
	rs = get_parm (s, ' ', ' ', 0);
	if (rs == 0)
	  goto erp;
	an = get_parm (rs, '\n', ' ', &delim);
	if (an == 0)
	  goto erp;
	if (delim == ' ') {
	  s = get_parm (an, '\n', '\n', 0);
	  if (s == 0)
		goto erp;
	} else
	  s = an;
#ifdef ARP_DAT
	if (f == 'A') {
	  if (delim == '\n')
		an = 0;
	  else
		an = expand_var (an, 0);
	  sold = expand_var (sold, &allocated);
	  if (sold == 0)
		return 0;
	  if (arp_add (sold, rs, an))
		return 0;
	  if(allocated)
		free (sold);
	  n++;
	} else
#endif
	if (f == 'J') {
	  if (delim == '\n')
		an = "*";
	  if (jump_add (sold, rs, an))
		return 0;
	  n++;
	} else if (f == 'M') {
	  if (delim == '\n')
		an = "*";
	  if (mirror_add (sold, rs, an))
		return 0;
	  n++;
	} else if (f == 'U') {
	  if (delim == '\n')
		an = star;
	  if (url_add (sold, rs, an)==0)
		return 0;
	  n++;
	} else if (f == 'P') {
	  if (delim == ' ')             /* P acl_paste acl1 acl2(+alcs...) */
		*(an - 1) = '+';        /* P acl_paste acl1+acl2(+alcs...) */
	  if (acl_paste (sold, rs))
		return 0;
	} else if (f == 'D' || f == 'R') {
	  int a2 = 0;

	  if (delim == '\n')
		an = "*";
	  else
		an = expand_var (an, &a2);
	  rs = expand_var (rs, &allocated);
	  if (an == 0 || rs == 0)
		return 0;
	  if (acl_add (f, sold, rs, an))
		return 0;
	  if(allocated)
		free (rs);
	  if (a2)
		free (an);
	} else if (f == 'S') {
	  if (delim == ' ') {
		if (rs[0] == '=' && rs[1] == 0)
		  rs = an;
		else {
		  print ("Error parsing cfg_file: S % %", rs, an);
		  return 0;
		}
	  }
	  if (def_var (sold, rs))
		return 0;
	} else if (f == 'L') {
	  if (delim == '\n')
		an = "U,C,-";
	  if (set_log_acl (sold, rs, an))
		return 0;
	} else if (f == 'I') {
	  if (delim == ' ') {
		print ("Error: not 2 arguments for I command");
		return 0;
	  }
	  f = *sold;
	  f &= 0337;
	  if ((f != 'E' && f != 'W') || *(sold + 1) != 0) {
		print ("Error parsing cfg_file: I % %", sold, rs);
		return 0;
	  }
	  rs = expand_var (rs, &allocated);
	  if (rs == 0 && f == 'E')
		return 0;
	  if(rs)
		f = read_ub (rs, 1);
	  if (rs != 0 && f == 0) {
		print ("%Error parsing include file %",
			   (f == 'W' ? "Warinig: " : ""), rs);
		if (f == 'E')
		  return 0;
	  }
	  if(rs && allocated)
		free (rs);
	  if(f > 0)
		n += f;
	} else if (f == 'F') {
	  char *buf, *s2;

	  if (delim == '\n')
		an = star;
	  sold = expand_var (sold, &allocated);
	  if (sold == 0)
		return 0;
	  if (stat (sold, &statb) < 0) {
		print ("Can`t stat url_file %", sold);
		return 0;
	  }
	  buf = my_alloc (statb.st_size + 2);
	  allocate2 = buf;

	  f = open (sold, O_RDONLY);
	  if (f < 0) {
		print ("Can`t open url_file %", sold);
		return 0;
	  }
	  s2 = buf + read (f, buf, statb.st_size);
	  close (f);
	  if ((s2 - buf) != statb.st_size ||
		   strip_space (buf, statb.st_size) == 0) {
		print ("Error reading url_file %", sold);
		return 0;
	  }
	  s2 = buf;
	  while (*s2) {
		while (*s2 != '\n' && *s2 != ' ')
		  s2++;
		*s2++ = 0;
		if (url_add (buf, rs, an)==0)
		  return 0;
		n++;
		buf = s2;
	  }
	  free (allocate2);
	  furl_add (sold, statb.st_mtime);
	  if(allocated)
		free (sold);
	}
  }
  if (n == 0) {
	if (recurs == 0)
	  print ("Error: parser config returned empty");
	else
	  n--;
  }
  free (allocate1);
  return n;
}


/*     */
#define SIZEBISIO 16384

static int ocnt;
static unsigned char *optr;

#define tstfd 0

static int
getc0 (void)
{
  static unsigned char obuf[SIZEBISIO];
  while (ocnt-- <= 0) {
	optr = obuf;
	if ((ocnt = read (tstfd, obuf, SIZEBISIO)) <= 0) {
#ifdef DEBUG
	  char buf[64];
	  buf[0] = 0;
	  put_timer (buf, ocnt);
	  print ("read returned %", buf);
	  buf[0] = 0;
	  put_timer (buf, errno);
	  print ("errno set to %", buf);
#endif
	  if (errno != EAGAIN && errno != EINTR)
		return ocnt;
	}
  }
  return (*optr++);
}

#define getc1()    (ocnt <= 0 ? getc0() : (ocnt--, *optr++))


static int
parse_buff (char *buf, char **src_addr, char **ident, char **method)
{
  char *s;
  int n, i;

  s = buf;
  while ((i = getc1 ()) > 0) {
	*s++ = i;
	if (i == '\n') {
	  *s = 0;
	  break;
	}
  }
  if (i <= 0)
	return i;
  s = buf;
  for (i = 0; i < 4; i++) {
	switch (i) {
	case 0:
	  break;
	case 1:
	  *src_addr = s;
	  break;
	case 2:
	  *ident = s;
	  break;
	default:
	  *method = s;
	  break;
	}
	n = 0;
	while (*s) {
	  if (*s == ' ' || *s == '\n') {
		*s = 0;
		s++;
		n++;
		break;
	  } else
		s++;
	}
	if (!n)
	  return 1;
  }
  return 2;
}

static int
match_and_not (const char *s, char *p)
{
  char *s0;
  char *ps;
  int m;

  s0 = p;
  while (*s0 != '\0') {
	if (*s0 == '^')
	  break;
	s0++;
  }
  if (*s0 == '\0')
	return match (p, s);
  *s0 = '\0';
  m = match (p, s);
  *s0 = '^';
  if (m == FNM_NOMATCH)
	return FNM_NOMATCH;
  s0++;
  for (ps = s0;; ps++) {
	  if(*ps == '\0') {
		if (s0 != ps && match (s0, s) == 0)
			return FNM_NOMATCH;
		return 0;
	  }
	  if (*ps == '^') {
		*ps = '\0';
		m = match (s0, s);
		*ps = '^';
		if (m == 0)
		  return FNM_NOMATCH;
		s0 = ps + 1;
	  }
  }
}

/*    glob_mask */
static
unsigned char de_code1[256] = {
  ['!' | 0200] = '!', ['(' | 0200] = '(', [')' | 0200] = ')',
  ['*' | 0200] = '*', ['+' | 0200] = '+', ['?' | 0200] = '?',
  [' ' | 0200] = '@', ['.' | 0200] = '[', [',' | 0200] = ']',
  ['-' | 0200] = '^' };
static void
decode_msk (unsigned char *msk)
{
  while (*msk) {
	if (de_code1[*msk])
	  *msk = de_code1[*msk];
	msk++;
  }
}

static char *
form_jump(char *url, const char *rurl, char *input_url)
{
    char *r = url;
    char sc;
    int m = FNM_NOMATCH;

    while(*++r) {
      sc = *r;
      *r = 0;
      m = match(rurl, url);
      *r = sc;
      if(m==0)
	break;
    }
    if(m==0) {
      char hex[4];
      char *endptr;

      input_url = r = input_url+(r-url);
      /*      %   */
      while (*r) {
	    if (*r == '%') {
	      if (*(r + 1) == '%')
		    strcpy (r + 1, r + 2);
	      else {
		    hex[0] = *(r + 1);
		    if (hex[0]) {
		      hex[1] = *(r + 2);
		      hex[2] = 0;
		      endptr = hex;
		      hex[3] = strtol (hex, &endptr, 16);
		      if (endptr != hex) {
			    *r = hex[3];
			    strcpy (r + 1, r + 1 + (endptr - hex));
		      }
		    }
	      }
	    }
	    r++;
      }
      if(strncasecmp (input_url, PREFIX_HTTP, PREFIX_HTTP_LEN) &&
	   strncasecmp (input_url, PREFIX_FTP, PREFIX_FTP_LEN)) {
		input_url -= PREFIX_HTTP_LEN;
		memcpy(input_url, PREFIX_HTTP, PREFIX_HTTP_LEN);
      }
    }
    return input_url;
}

static char *
form_mirror(char *url, struct LIST_U_RU *cur, char *input_url)
{
    char *r = url;
    char *u0 = cur->url;
    size_t l = strlen(u0) - 1;
    char sc;
    int m = FNM_NOMATCH;

    u0[l] = '\0';       /* tail '*' -> '\0' */
    while(*++r) {
      sc = *r;
      *r = 0;
      m = match(u0, url);
      *r = sc;
      if(m == 0)
	break;
    }
    u0[l] = '*';       /* restore '*' */
    if(m == 0) {
	strcpy(url, cur->rurl);
	strcat(url, input_url + (r-url));
	return url;
    }
    return input_url;
}

/*     */
static void
print_spec (const char *url, const char *src_addr,
		const char *ident, const char *method, int flg_moved)
{
  char *buf, *pbuf;
  int lu, ls, li, lm;

  if(flg_moved)
    flg_moved = 4;  /* strlen("302:") - Moved temporarity */
  lu = strlen (url);
  ls = strlen (src_addr);
  li = strlen (ident);
  lm = strlen (method);
  buf = alloca (flg_moved + lu + ls + li + lm + 4); /* + 3   \n */

  if(flg_moved)
    strcpy(buf, "302:");  /* HTTP Code: Moved temporarity */
  pbuf = strcpy (buf+ flg_moved, url) + lu;
  *pbuf++ = ' ';
  pbuf = strcpy (pbuf, src_addr) + ls;
  *pbuf++ = ' ';
  pbuf = strcpy (pbuf, ident) + li;
  *pbuf++ = ' ';
  pbuf = strcpy (pbuf, method) + lm;
  *pbuf++ = '\n';
  write (1, buf, (pbuf - buf));
}


static volatile long tm_ma;
static volatile int  tm_wday;

static int
time_is_cur (const struct ALC_STRUCT *dim_acl)
{
  int i;

  for (i = 0; i < dim_acl->ntime; i++)
	if (tm_wday >= dim_acl->time[i].d1 &&
		tm_wday <= dim_acl->time[i].d2 &&
		tm_ma >= dim_acl->time[i].min1 &&
		tm_ma <= dim_acl->time[i].min2)
	  return 1;
  return 0;
}


#include <signal.h>

/* start load and reload cfg */
static volatile int signaled = -1;

static void
sig_usr1 (int s)
{
#if defined(SA_RESTART) || defined(SA_INTERRUPT)
  struct sigaction sa;
#endif

  signaled = 0;
  signal (SIGUSR1, sig_usr1);
#if defined(SA_RESTART) || defined(SA_INTERRUPT)
  sigaction (SIGUSR1, (struct sigaction *) 0, &sa);
#ifdef SA_RESTART
  sa.sa_flags |= SA_RESTART;
#endif
#ifdef SA_INTERRUPT
  sa.sa_flags &= ~SA_INTERRUPT;
#endif
  sigaction (SIGUSR1, &sa, (struct sigaction *) 0);
#endif
}


static time_t timer;

#ifdef ARP_DAT
/* need reload the kernel arp table, reloading make after alarm time */
static int need_reload_arp = 0;
static volatile signed char arp_signaled = -1;
#endif

static void
sig_alarm (int s)
{
  struct tm *tms;
  struct stat statb;
  struct F_URLS_REREAD *cur;

#if defined(SA_RESTART) || defined(SA_INTERRUPT)
  struct sigaction sa;
#endif

  signal (SIGALRM, sig_alarm);
#if defined(SA_RESTART) || defined(SA_INTERRUPT)
  sigaction (SIGALRM, (struct sigaction *) 0, &sa);
#ifdef SA_RESTART
  sa.sa_flags |= SA_RESTART;
#endif
#ifdef SA_INTERRUPT
  sa.sa_flags &= ~SA_INTERRUPT;
#endif
  sigaction (SIGALRM, &sa, (struct sigaction *) 0);
#endif

#ifdef ARP_DAT
  /* reload by alarm if need */
  if(need_reload_arp) {
	need_reload_arp = 0;
	arp_signaled = 1;
  }
#endif
  timer = time (NULL);
  tms = localtime (&timer);
  tm_ma =
	(long) tms->tm_sec + tms->tm_min * 60 + (long) tms->tm_hour * 60 * 60;
  tm_wday = tms->tm_wday;
  for (cur = top_fur; cur != 0; cur = cur->next)
	if (stat (cur->name, &statb) == 0)
	  if (statb.st_mtime > cur->st_m) {
		signaled = 0;
		return;
	  }
  alarm (ALARM_TIME);
}


#ifdef ARP_DAT

static void
sig_arp (int s)
{

#if defined(SA_RESTART) || defined(SA_INTERRUPT)
  struct sigaction sa;
#endif

  signal (SIGUSR2, sig_arp);
#if defined(SA_RESTART) || defined(SA_INTERRUPT)
  sigaction (SIGUSR2, (struct sigaction *) 0, &sa);
#ifdef SA_RESTART
  sa.sa_flags |= SA_RESTART;
#endif
#ifdef SA_INTERRUPT
  sa.sa_flags &= ~SA_INTERRUPT;
#endif
  sigaction (SIGUSR2, &sa, (struct sigaction *) 0);
#endif

  arp_signaled = 1;
}

static void
decode_arp_dat (void)
{
  struct ARP_LIST_CMP *cur;

  char arp[6];
  char buf[80];
  FILE *pf;
  unsigned int fi;
  char *s, *e;
  int fl = 0;

#ifdef DEBUG
  print ("reload arp table");
#endif
  arp_signaled = 0;
  pf = fopen (arp_dat, "r");
  if (pf == NULL) {
	print ("Can`t open %", ARP_DAT);
	return;                                 /* ignore error */
  }
  while (fgets (buf, sizeof (buf), pf) != NULL) {
#ifdef DEBUG
	print ("buf = %", buf);
#endif
	if(fl == 0) {
		/*    */
		fl++;
		continue;
	}
#ifdef DEBUG
	print ("buf before scan_ip = %", buf);
#endif
	s = buf;
	/*
	   format:
	   1.2.3.4 0x1 0x2 00:01:02:03:04:05 * eth0
	*/
	while(*s != ' ' && *s != '\t' && *s)
		s++;
	if(*s == 0)
		continue;       /* hmm */
	*s = '/';
	s = buf;
	if (scan_ip (&s, &fi) != 1)
	  continue;
	if(strtol(s, &e, 16) != 1)
	  continue;
	s = e;
	if(strtol(s, &e, 16) != 2)
	  continue;
	s = e;
	while(*s == ' ' || *s == '\t')
		s++;
	if(*s == 0)
		continue;       /* hmm */
	if(decode_arp (s, arp) != 0)
	  continue;
	if(arp[0] == 0 && arp[1] == 0 && arp[2] == 0 && arp[3] == 0
		&& arp[4] == 0 && arp[5] == 0) {
#ifdef DEBUG
	  print ("found incomplete arp");
#endif
	  continue;
	}
	for (cur = arp_cmp_top; cur != 0; cur = cur->next)
		if (cur->ip == fi) {
#ifdef DEBUG
		  if (memcmp (cur->arp, arp, 6))
			print ("arp changed");
#endif
		  memcpy (cur->arp, arp, 6);
		  break;
		}
	if (cur == 0) {
		cur = my_alloc (sizeof (struct ARP_LIST_CMP));
		if (arp_cmp_top == 0)
		  arp_cmp_top = cur;
		else {
		  struct ARP_LIST_CMP *b = arp_cmp_top;
		  while (b->next)
			b = b->next;
		  b->next = cur;
		}
		cur->ip = fi;
		memcpy (cur->arp, arp, 6);
		cur->next = 0;
#ifdef DEBUG
		print ("new arp loaded");
#endif
	}
  }
  fclose (pf);
#ifdef DEBUG
  if (arp_cmp_top == 0)
	print("arp table empty");
  else
	print("arp loaded");
#endif
}

#endif


static void
put_log (int log_type, char *input_url, char *src_addr,
		 char *ident, char *acl_name, char *out_url, char *msk)
{
  char *out_str;
  char tl[32];
  int fd, l;

  l = strlen (input_url) + strlen (src_addr) + strlen (ident) +
	strlen (acl_name) + strlen (out_url) + strlen (msk) + 6;
  if ((log_type & LOG_OPT_D) != 0) {
	tl[0] = 0;
	put_timer (tl, timer);
	l += strlen (tl) + 1;
  }
  out_str = alloca (l + 1);
  if (out_str == 0)
	return;
  if ((log_type & LOG_OPT_D) != 0) {
	strcpy (out_str, tl);
	strcat (out_str, " ");
  } else
	out_str[0] = 0;
  if ((log_type & LOG_OPT_U) != 0) {
	strcat (out_str, input_url);
	strcat (out_str, " ");
  }
  if ((log_type & LOG_OPT_C) != 0) {
	strcat (out_str, src_addr);
	strcat (out_str, "/");
	strcat (out_str, ident);
	strcat (out_str, " ");
  }
  if ((log_type & LOG_OPT_G) != 0) {
	char *old;
	old = &out_str[strlen (out_str)];
	strcpy (old, msk);
	decode_msk ((unsigned char *)old);
	strcat (out_str, " ");
  }
  if ((log_type & LOG_OPT_N) != 0) {
	strcat (out_str, acl_name);
	strcat (out_str, " ");
  }
  if ((log_type & LOG_OPT_O) != 0 ||
	  ((log_type & (LOG_OPT_R | LOG_OPT_M)) != 0 && input_url != out_url))
	strcat (out_str, out_url);
  if ((log_type & LOG_OPT_M) != 0 && input_url == out_url)
	strcat (out_str, "-");
  l = strlen (out_str);
  if (out_str[l - 1] == ' ')
	out_str[l - 1] = '\n';
  else
	out_str[l++] = '\n';
  fd = open (REDIR_LOG_NAME, O_WRONLY | O_CREAT | O_APPEND, 0660);
  if (fd > 0) {
	flock (fd, LOCK_EX);    /* squid started not 1 redirector process */
	write (fd, out_str, l);
	close (fd);
  }
}



#ifdef DEBUG
#define CORE_SIZE 1024000
#include <sys/resource.h>
#endif

static void
coding_client_url (char *gu)
{
#ifndef PERCENT_HEX_HIDE
  char hex[4];
  char *endptr = gu;
#endif

  while (*gu) {
	*gu = code[(unsigned char) *gu];
	gu++;
  }
#ifndef PERCENT_HEX_HIDE
  gu = endptr;
  while (*gu) {
	if (*gu == '%') {
	  if (*(gu + 1) == '%')
		strcpy (gu + 1, gu + 2);
	  else {
		hex[0] = *(gu + 1);
		if (hex[0]) {
		  hex[1] = *(gu + 2);
		  hex[2] = 0;
		  endptr = hex;
		  hex[3] = strtol (hex, &endptr, 16);
		  if (endptr != hex) {
			*gu = code[(unsigned char) hex[3]];
			strcpy (gu + 1, gu + 1 + (endptr - hex));
		  }
		}
	  }
	}
	gu++;
  }
#endif
}

#ifdef ARP_DAT
static void
cmp_arp (unsigned client_ip, char *input_url, char *src_addr, char *ident,
		 char **rurl)
{

  struct ARP_LIST *cur;
  struct ARP_LIST *cur0 = NULL;
  struct ARP_LIST_CMP *cur_cmp = NULL;

  for (cur = arp_top; cur != 0; cur = cur->next)
	if (client_ip == cur->ip) {
	  /*   MAC    */
	  if (cur_cmp == NULL) {
		for (cur_cmp = arp_cmp_top; cur_cmp != NULL; cur_cmp = cur_cmp->next)
			if (client_ip == cur_cmp->ip)
				break;
		if (cur_cmp == NULL) {
			/* Hmm,    (  arp -a) */
			need_reload_arp++;
#ifdef DEBUG
			print("client ip presend not from arps defines?");
#endif
			return;         /*    */
		}
	  }
	  if (memcmp (cur->arp, cur_cmp->arp, 6) == 0)
		return;                                 /*  */
	  if (cur0 == NULL)
		cur0 = cur;
	}
  if (cur0 == NULL)     /*  IP */
	return;
  /*  ,    A  */
  *rurl = cur0->rurl;
  if (cur0->log_acl != 0)
	put_log (cur0->log_acl->dim_acl[0].log_type, input_url,
			 src_addr, ident, "ARP_NO_MATCH", cur0->rurl, "*");
}
#endif

	/*       ;) */
#define acl_log cur->acl_name->dim_acl[i].log_type
#define LOG_OPT_TRA (LOG_OPT_TR | LOG_OPT_TA)
#define LOG_OPT_TDA (LOG_OPT_TD | LOG_OPT_TA)

int
main (int argc, char **argv)
{
  char *src_addr = src_addr, *ident = ident, *method = method; /* make gcc happy */
  int i;
  char buf[SIZEBISIO];
  char glob_url[SIZEBISIO];
  char *code_glob;
  unsigned client_ip;
  struct LIST_U_RU *cur;
#ifdef DEBUG
  struct rlimit rlim;
  getrlimit (RLIMIT_CORE, &rlim);
  if (rlim.rlim_max < CORE_SIZE)
	rlim.rlim_cur = rlim.rlim_max;
  else
	rlim.rlim_cur = CORE_SIZE;
  setrlimit (RLIMIT_CORE, &rlim);
#endif

  if (argc > 2) {
	print ("usage: % [cfg_file]", argv[0]);
	return 1;
  }
  if (argc != 2) {
	if (chdir (CHDIR_NAME) < 0) {
	  print ("Can`t chdir %", CHDIR_NAME);
	  return 1;
	}
  }
  signal (SIGTERM, SIG_IGN);
  i = 0;
  for (;;) {
	if (signaled <= 0) {
	  if (signaled < 0) {
		sig_usr1 (SIGUSR1);
#ifdef ARP_DAT
		sig_arp (SIGUSR2);
#else
		signal (SIGUSR2, SIG_IGN);
#endif
	  } else {
		alarm (0);
		signal (SIGALRM, SIG_IGN);
		/* reset allocated */
		free_all_lists ();
	  }
#ifdef DEBUG
	  print ("DEBUG: start config parser");
#endif
	  if (!read_ub (argc == 2 ? argv[1] : REDIR_CFG_NAME, 0))
		return 1;
#ifdef DEBUG
	  print ("DEBUG: end config parser");
#endif
	  sig_alarm (SIGALRM);
	  signaled++;
	}
#ifdef ARP_DAT
	if (arp_signaled > 0)
	  decode_arp_dat ();
#endif
	if (i == 0) {
	  i = parse_buff (buf, &src_addr, &ident, &method);
	  if (i != 2) {
		if (i < 0)
		  print ("Parser returned < 0");
#ifdef DEBUG
		else if (i == 0)
		  print ("stdin EOF");
#endif
		else if (i == 1)
		  print ("Error: bad format stdin input");
		return i < 0 ? -i : i;
	  }
	  continue;     /*        */
	}
#ifdef DEBUG
	print("url=% src_addr=% ident=% method=%", buf, src_addr, ident, method);
#endif
	coding_client_url (strcpy (glob_url, buf));
	code_glob = src_addr;
	if (scan_ip (&code_glob, &client_ip) <= 0) {
	  print ("Bad decode client address");
	  return 2;
	}
	code_glob = buf;
	cur = strcmp(method, "CONNECT") == 0 ? top_connect : top;
	for (; cur; cur = cur->next) {
	  if (match_and_not (glob_url, cur->url)==0) {
		for (i = 0; i < cur->acl_name->n; i++)
		  if ((client_ip & cur->acl_name->dim_acl[i].mask) ==
			  cur->acl_name->dim_acl[i].ip) {
			if (time_is_cur (&cur->acl_name->dim_acl[i])) {
			  if (cur->acl_name->dim_acl[i].type == 'R') {
				code_glob = cur->rurl;
				if ( (acl_log & LOG_OPT_TRA) != 0)
				  i = acl_log;
				else
				  i = 0;
			  } else {
				if ( (acl_log & LOG_OPT_TDA) != 0)
				  i = acl_log;
				else
				  i = 0;
			  }
			  if(code_glob != buf && cur->jump_mirror_connect == 1) {
			    if(cur->jump_mirror_connect == 1)
			      code_glob = form_jump(glob_url, cur->rurl, buf);
			    else
			      code_glob = form_mirror(glob_url, cur, buf);
			  }
			  if (i)
				put_log (i, buf, src_addr, ident,
						cur->acl_name->name,
						 code_glob, cur->url);
			  goto break_loop;
			}
		  }
	  }
	}
  break_loop:
#ifdef ARP_DAT
	if (code_glob == buf)
	  cmp_arp (client_ip, buf, src_addr, ident, &code_glob);
#endif
	print_spec (code_glob, src_addr, ident, method, cur!=0 && cur->jump_mirror_connect==1);
	i = 0;
  }
}
